/////////////////////////////////////////////////////////////////////////////////////
// Katax Emperore  -- web3D sample  --
// Web3D + JavaScript programming by using three.js
// for more samples and details, please check my home page-->   www.kataxwork.com
/////////////////////////////////////////////////////////////////////////////////////
// This application makes number of shapes (like circle, star), out of 10 cube 3d-elements on stage.
// During play time, elements have self-rotations, and after a while, all elements move
// to different coordinations to make a different shape. The program, repeat the cycle for
// all 5 shapes which are defined in “Shapes” class.
// Also, back ground has very low speed rotation.
/////////////////////////////////////////////////////////////////////////// 2017 ////

// scene and render
var areaWidth           = document.getElementById('web3d').offsetWidth; // aspect ratio
var areaHeight          = document.getElementById('web3d').offsetHeight;// aspect ratio
var renderer;
var scene;
var camera;
var cameraControl;

// elements on the stage
var shapeBank           = [];   // element`s object holder
var shapeName           = [];   // values: ["circle", "landa_horizonal", "cross", "landa_vertical", "star" ];
var shapeMax;
var shapeIterator       = 4;                // use for cycling between shapes
var shapeCoords         = new Shapes();     // shape coordinations and names
var sky                 = new Sky();        // sky element

// move
var currentPos          = new THREE.Vector3();  // current element`s coordination
var currentPosCounter   = 0;                    // counter for control element`s movement
var nextPos             = new THREE.Vector3();  // next element`s coordinatoion
var dist                = 0;                    // distance between cur and next coordination of each element
var distBuffer          = [];                   // buffer distance
var angle               = 0;                    // angle of element`s movement
var moveSpeed           = .2;                   // speed of element`s movement
var moveFlag            = false;                // control element`s movement
var timeOutValue        = 1480;                 // timeout between each shape transitions

// helper
var outConsoleCounter   = 0;                    // times to print message in console (F 12)


init();

////////////////////////////////// render & camera  //////////

function create_Renderer() {
    renderer = new THREE.WebGLRenderer();
    renderer.setClearColor(0x000000, 1.0);
    renderer.setSize(areaWidth, areaHeight);
    renderer.shadowMap.enabled = true;
}

function create_Camera() {
    camera = new THREE.PerspectiveCamera(
        45,
        areaWidth / areaHeight,
        0.1, 1000);
    camera.position.x = 10;
    camera.position.y = -10;
    camera.position.z = -10;
    camera.lookAt(scene.position);
    cameraControl = new THREE.OrbitControls(camera);
}

///////////////////////////////////////////// Light & Sky /////////

function create_SpotLight() {
    var spotLight = new THREE.SpotLight(0xffffff);
    spotLight.position.set(10, 20, 20);
    spotLight.shadowCameraNear = 20;
    spotLight.shadowCameraFar = 50;
    spotLight.castShadow = true;
    scene.add(spotLight);
}

function create_DirectionalLight() {
    var directionalLight = new THREE.DirectionalLight(0xffffff, 1);
    directionalLight.position.set(100, 10, -50);
    //directionalLight.name = "directional";
    var ambientLight = new THREE.AmbientLight(0x555555);

    scene.add(directionalLight);
    scene.add(ambientLight);
}

function create_Sky() {     scene.add(sky.SkyMesh);     }

////////////////////////////////////////// elements on stage //////

function create_Object() {
    shapeName = shapeCoords.getShapeNames();    // get shape names
    shapeMax = shapeName.length;                // get max number of shapes
    setScalarValues();                          // rescale and set coordination matrix for elements
    var shapename = shapeName[shapeIterator];   // shape name
    var shapeX = shapeCoords.getCoordX(shapename);  // extract coordinations of elements for X
    var shapeY = shapeCoords.getCoordY(shapename);  // extract coordinations of elements for Y
    for (var i = 0; i < shapeCoords.getMaxObjects(); i++){      // creat elements and put on stage
        shapeBank[i] = new Cube();
        shapeBank[i].setCoordinations(shapeX[i], shapeY[i], 0);
        shapeBank[i].setScale(.1, .1, .1);
        scene.add(shapeBank[i].CubeMesh);
        distBuffer[i] = 1000;                       // set distant buffer array for first run
    }

}

function setScalarValues() {                     // rescale and set coordination matrix for elements
    for (var i = 0; i < shapeMax; i++){
        shapeCoords.setShiftPositions(shapeName[i], -6, -6, 0);
        shapeCoords.setScaleMultiplier(shapeName[i], 1);
    }
}

////////////////////////////////////////// transitions //////////
function rotator() {
    for (var i = 0; i < shapeCoords.getMaxObjects(); i++) shapeBank[i].CubeMesh.rotation.y += 0.007 + (i * .0003);
    sky.SkyMesh.rotation.y += .0001;
}

function changeShape(){
    shapeIterator++;
    if (shapeIterator == shapeMax) shapeIterator = 0;
    // load distBuffer with 1000 (very huge number).... this shows movmemnt is ready to go for object i
    for (var i = 0; i < shapeBank.length; i++) distBuffer[i] = 1000;
    moveFlag = true;
    //outConsole(" B--- --  "  + " > <  " + shapeIterator);
}


function mover(){
    if (moveFlag) {
        for ( var i = 0; i < shapeBank.length; i++ ) {
            if ( distBuffer[i] != -1 ) {
                currentPos = shapeBank[i].getCoordinations();
                nextPos.x = shapeCoords.getX(shapeName[shapeIterator], i);
                nextPos.y = shapeCoords.getY(shapeName[shapeIterator], i);
                if ( ( currentPos.x != nextPos.x ) || (currentPos.y != nextPos.y) ) {
                    dist = Math.sqrt(( currentPos.x - nextPos.x ) * ( currentPos.x - nextPos.x ) + ( currentPos.y - nextPos.y ) * ( currentPos.y - nextPos.y ));
                    if ( distBuffer[i] > dist ) {
                        angle = Math.atan2(( nextPos.y - currentPos.y ), ( nextPos.x - currentPos.x ));// shoot angle
                        currentPos.x += ( moveSpeed * Math.cos(angle) );
                        currentPos.y += ( moveSpeed * Math.sin(angle) );
                        distBuffer[i] = dist;
                    } else {
                        currentPos.x = nextPos.x;
                        currentPos.y = nextPos.y;
                    }
                    shapeBank[i].setCoordinations(currentPos.x, currentPos.y, 0);
                } else {
                    if ( distBuffer[i] != -1 ) currentPosCounter++;
                    // load distBuffer with -1 ... this shows movmemnt is done for object i
                    distBuffer[i] = -1;
                }
            }
        }
    }
    if (currentPosCounter == shapeCoords.getMaxObjects() ){
        currentPosCounter = 0;
        moveFlag = false;
        setTimeout(changeShape, timeOutValue); // change shape at the end of move
    }
}


//////////////////////////////// strat init //////////////////////
function init() {
    scene = new THREE.Scene();

    create_Renderer();
    create_Camera();
    //create_SpotLight();
    create_DirectionalLight();
    create_Sky();
    create_Object();

    changeShape(); // change shape on first run

    document.getElementById('web3d').appendChild(renderer.domElement);
    render(); // loop function
}

////////////////////////////////  engine loop /////////////////////
function render() {
    cameraControl.update();
    rotator();
    mover();
    renderer.render(scene, camera);
    requestAnimationFrame(render);
}


////////////////////////////////  helper /////////////////////
function outConsole(txt){  // print output on console with limited number set to 100
    if (outConsoleCounter <= 100) {
        outConsoleCounter++;
        console.log("\n" + txt + "\n");
    }
}


